/* QLogic (R)NIC Driver/Library
 * Copyright (c) 2015-2016  QLogic Corporation
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <linux/types.h>
#include <linux/string.h>
#define __PREVENT_DUMP_MEM_ARR__
#define __PREVENT_PXP_GLOBAL_WIN__
#include "qed.h"
#include "qed_dev_api.h"
#include "qed_hw.h"
#include "qed_l2.h"
#include "qed_mcp.h"
#include "qed_ptp_api.h"
#include "qed_reg_addr.h"

/* 16 nano second time quantas to wait before making a Drift adjustment */
#define QED_DRIFT_CNTR_TIME_QUANTA_SHIFT    0
/* Nano seconds to add/subtract when making a Drift adjustment */
#define QED_DRIFT_CNTR_ADJUSTMENT_SHIFT     28
/* Add/subtract the Adjustment_Value when making a Drift adjustment */
#define QED_DRIFT_CNTR_DIRECTION_SHIFT      31

static int qed_ptp_res_lock(struct qed_hwfn *p_hwfn, struct qed_ptt *p_ptt)
{
	struct qed_resc_lock_params params;
	int rc;

	memset(&params, 0, sizeof(params));
	switch (qed_device_get_port_id(p_hwfn->cdev)) {
	case 0:
		params.resource = QED_RESC_LOCK_PTP_PORT0;
		break;
	case 1:
		params.resource = QED_RESC_LOCK_PTP_PORT1;
		break;
	case 2:
		params.resource = QED_RESC_LOCK_PTP_PORT2;
		break;
	case 3:
		params.resource = QED_RESC_LOCK_PTP_PORT3;
		break;
	default:
		return -EINVAL;
	}

	params.timeout = QED_MCP_RESC_LOCK_TO_NONE;
	rc = qed_mcp_resc_lock(p_hwfn, p_ptt, &params);
	if (rc != 0 && rc != -EINVAL) {
		return rc;
	} else if (rc == -EINVAL) {
		/* MFW doesn't support resource locking, first PF on the port
		 * has lock ownership.
		 */
		if (p_hwfn->abs_pf_id < p_hwfn->cdev->num_ports_in_engine) {
			return 0;
		} else {
			DP_INFO(p_hwfn, "PF doesn't have lock ownership\n");
			return -EBUSY;
		}
	} else if (rc == 0 && !params.b_granted) {
		DP_INFO(p_hwfn, "Failed to acquire ptp resource lock\n");
		return -EBUSY;
	}

	return rc;
}

static int qed_ptp_res_unlock(struct qed_hwfn *p_hwfn, struct qed_ptt *p_ptt)
{
	struct qed_resc_unlock_params params;
	int rc;

	memset(&params, 0, sizeof(params));
	switch (qed_device_get_port_id(p_hwfn->cdev)) {
	case 0:
		params.resource = QED_RESC_LOCK_PTP_PORT0;
		break;
	case 1:
		params.resource = QED_RESC_LOCK_PTP_PORT1;
		break;
	case 2:
		params.resource = QED_RESC_LOCK_PTP_PORT2;
		break;
	case 3:
		params.resource = QED_RESC_LOCK_PTP_PORT3;
		break;
	default:
		return -EINVAL;
	}

	rc = qed_mcp_resc_unlock(p_hwfn, p_ptt, &params);
	if (rc == -EINVAL) {
		/* MFW doesn't support locking, first PF has lock ownership */
		if (p_hwfn->abs_pf_id < p_hwfn->cdev->num_ports_in_engine) {
			rc = 0;
		} else {
			DP_INFO(p_hwfn, "PF doesn't have lock ownership\n");
			return -EINVAL;
		}
	} else if (rc) {
		DP_INFO(p_hwfn, "Failed to release the ptp resource lock\n");
	}

	return rc;
}

int qed_ptp_hwtstamp_tx_on(struct qed_hwfn *p_hwfn, struct qed_ptt *p_ptt)
{
	qed_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_PARAM_MASK, 0x4A0);
	qed_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_RULE_MASK, 0x3CAA);

	return 0;
}

int qed_ptp_disable(struct qed_hwfn *p_hwfn, struct qed_ptt *p_ptt)
{
	qed_ptp_res_unlock(p_hwfn, p_ptt);

	/* Reset PTP event detection rules */
	qed_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_PARAM_MASK, 0x7FF);
	qed_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_RULE_MASK, 0x3FFF);

	qed_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_PARAM_MASK, 0x7FF);
	qed_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_RULE_MASK, 0x3FFF);

	/* Disable the PTP feature */
	qed_wr(p_hwfn, p_ptt, NIG_REG_RX_PTP_EN, 0x0);
	qed_wr(p_hwfn, p_ptt, NIG_REG_TX_PTP_EN, 0x0);

	return 0;
}

/* Read Rx timestamp */
int qed_ptp_read_rx_ts(struct qed_hwfn *p_hwfn,
		       struct qed_ptt *p_ptt, u64 * timestamp)
{
	u32 val;

	*timestamp = 0;
	val = qed_rd(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_SEQID);
	if (!(val & 0x10000)) {
		DP_INFO(p_hwfn, "Invalid timestamp\n");
		return -EINVAL;
	}

	val = qed_rd(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_TS_LSB);
	*timestamp = qed_rd(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_TS_MSB);
	*timestamp <<= 32;
	*timestamp |= val;

	/* Reset timestamp register to allow new timestamp */
	qed_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_SEQID, 0x10000);

	return 0;
}

/* Read Tx timestamp */
int qed_ptp_read_tx_ts(struct qed_hwfn *p_hwfn,
		       struct qed_ptt *p_ptt, u64 * timestamp)
{
	u32 val;

	*timestamp = 0;
	val = qed_rd(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_BUF_SEQID);
	if (!(val & 0x10000))
		return -EINVAL;

	val = qed_rd(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_BUF_TS_LSB);
	*timestamp = qed_rd(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_BUF_TS_MSB);
	*timestamp <<= 32;
	*timestamp |= val;

	/* Reset timestamp register to allow new timestamp */
	qed_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_BUF_SEQID, 0x10000);

	return 0;
}

/* Read Phy Hardware Clock */
int qed_ptp_read_cc(struct qed_hwfn *p_hwfn,
		    struct qed_ptt *p_ptt, u64 * phc_cycles)
{
	u32 temp = 0;

	temp = qed_rd(p_hwfn, p_ptt, NIG_REG_TSGEN_SYNC_TIME_LSB);
	*phc_cycles = qed_rd(p_hwfn, p_ptt, NIG_REG_TSGEN_SYNC_TIME_MSB);
	*phc_cycles <<= 32;
	*phc_cycles |= temp;

	return 0;
}

/* Filter PTP protocol packets that need to be timestamped */
int qed_ptp_cfg_rx_filters(struct qed_hwfn *p_hwfn,
			   struct qed_ptt *p_ptt, enum qed_ptp_filter_type type)
{
	u32 rule_mask, parm_mask;

	switch (type) {
	case QED_PTP_FILTER_L2_IPV4_IPV6:
		parm_mask = 0x4A0;
		rule_mask = 0x3CAA;
		break;
	case QED_PTP_FILTER_L2:
		parm_mask = 0x4BF;
		rule_mask = 0x3CFF;
		break;
	case QED_PTP_FILTER_IPV4_IPV6:
		parm_mask = 0x7E0;
		rule_mask = 0x3FFA;
		break;
	case QED_PTP_FILTER_IPV4:
		parm_mask = 0x7EC;
		rule_mask = 0x3FFA;
		break;
	default:
		DP_INFO(p_hwfn, "Invalid PTP filter type %d\n", type);
		return -EINVAL;
	}

	qed_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_PARAM_MASK, parm_mask);
	qed_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_RULE_MASK, rule_mask);

	/* Reset possibly old timestamps */
	qed_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_SEQID, 0x10000);

	return 0;
}

int qed_ptp_adjfreq(struct qed_hwfn *p_hwfn, struct qed_ptt *p_ptt, s32 ppb)
{
	s64 period, period1, period2, dif, dif1, dif2;
	int drift_dir, best_val, best_period;
	s64 best_dif, temp, val;
	u32 drift_ctr_cfg = 0;
	u32 drift_state;

	best_dif = 1000000000;
	best_period = 1;
	best_val = 0;
	drift_dir = 1;

	if (ppb < 0) {
		ppb = -ppb;
		drift_dir = 0;
	}

	if (ppb == 0) {
		best_val = 0;
		best_period = 0xFFFFFFF;
	} else {
		for (val = 0; val <= 7; val++) {
			period1 = div64_s64(val * 1000000000, (s64) ppb);
			period1 = div64_s64(period1 - 8, (s64) 16);
			if (period1 < 1)
				period1 = 1;
			if (period1 > 0xFFFFFFE)
				period1 = 0xFFFFFFE;
			period2 = period1 + 1;

			temp = div64_s64(val * 1000000000, period1 * 16 + 8);
			dif1 = ppb - temp;
			if (dif1 < 0)
				dif1 = -dif1;

			temp = div64_s64(val * 1000000000, period2 * 16 + 8);

			dif2 = ppb - temp;
			if (dif2 < 0)
				dif2 = -dif2;

			dif = min_t(s64, dif1, dif2);
			period = (dif1 < dif2) ? period1 : period2;
			if (dif < best_dif) {
				best_dif = dif;
				best_val = (int)val;
				best_period = (int)period;
			}
		}
	}

	drift_ctr_cfg = (best_period << QED_DRIFT_CNTR_TIME_QUANTA_SHIFT) |
	    (best_val << QED_DRIFT_CNTR_ADJUSTMENT_SHIFT) |
	    (drift_dir << QED_DRIFT_CNTR_DIRECTION_SHIFT);

	qed_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_RST_DRIFT_CNTR, 0x1);

	drift_state = qed_rd(p_hwfn, p_ptt, NIG_REG_TSGEN_RST_DRIFT_CNTR);
	if (drift_state & 1) {
		qed_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_DRIFT_CNTR_CONF,
		       drift_ctr_cfg);
	} else {
		DP_INFO(p_hwfn, "Drift counter is not reset\n");
		return -EINVAL;
	}

	qed_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_RST_DRIFT_CNTR, 0x0);

	return 0;
}

int qed_ptp_enable(struct qed_hwfn *p_hwfn, struct qed_ptt *p_ptt)
{
	int rc;

	rc = qed_ptp_res_lock(p_hwfn, p_ptt);
	if (rc) {
		DP_INFO(p_hwfn,
			"Couldn't acquire the resource lock, skip ptp enable for this PF\n");
		return rc;
	}

	/* Reset PTP event detection rules - will be configured in the IOCTL */
	qed_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_PARAM_MASK, 0x7FF);
	qed_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_RULE_MASK, 0x3FFF);
	qed_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_PARAM_MASK, 0x7FF);
	qed_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_RULE_MASK, 0x3FFF);

	qed_wr(p_hwfn, p_ptt, NIG_REG_TX_PTP_EN, 7);
	qed_wr(p_hwfn, p_ptt, NIG_REG_RX_PTP_EN, 7);

	qed_wr(p_hwfn, p_ptt, NIG_REG_TS_OUTPUT_ENABLE_PDA, 0x1);

	/* Pause free running counter */
	if (QED_IS_BB_B0(p_hwfn->cdev))
		qed_wr(p_hwfn, p_ptt, NIG_REG_TIMESYNC_GEN_REG_BB, 2);
	if (QED_IS_AH(p_hwfn->cdev) || QED_IS_E5(p_hwfn->cdev))
		qed_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_FREECNT_UPDATE_K2_E5, 2);

	qed_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_FREE_CNT_VALUE_LSB, 0);
	qed_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_FREE_CNT_VALUE_MSB, 0);
	/* Resume free running counter */
	if (QED_IS_BB_B0(p_hwfn->cdev))
		qed_wr(p_hwfn, p_ptt, NIG_REG_TIMESYNC_GEN_REG_BB, 4);
	if (QED_IS_AH(p_hwfn->cdev) || QED_IS_E5(p_hwfn->cdev)) {
		qed_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_FREECNT_UPDATE_K2_E5, 4);
		/* Workaround for latching the 2-step PTP packets */
		qed_wr(p_hwfn, p_ptt, NIG_REG_PTP_LATCH_OSTS_PKT_TIME_K2_E5, 1);
	}

	/* Disable drift register */
	qed_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_DRIFT_CNTR_CONF, 0x0);
	qed_wr(p_hwfn, p_ptt, NIG_REG_TSGEN_RST_DRIFT_CNTR, 0x0);

	/* Reset possibly old timestamps */
	qed_wr(p_hwfn, p_ptt, NIG_REG_LLH_PTP_HOST_BUF_SEQID, 0x10000);
	qed_wr(p_hwfn, p_ptt, NIG_REG_TX_LLH_PTP_BUF_SEQID, 0x10000);

	return 0;
}

static int qed_ptp_hw_hwtstamp_tx_on(struct qed_dev *cdev)
{
	struct qed_hwfn *hwfn = QED_LEADING_HWFN(cdev);

	return qed_ptp_hwtstamp_tx_on(hwfn, hwfn->p_ptp_ptt);
}

static int qed_ptp_hw_cfg_rx_filters(struct qed_dev *cdev,
				     enum qed_ptp_filter_type type)
{
	struct qed_hwfn *hwfn = QED_LEADING_HWFN(cdev);

	return qed_ptp_cfg_rx_filters(hwfn, hwfn->p_ptp_ptt, type);
}

static int qed_ptp_hw_read_rx_ts(struct qed_dev *cdev, u64 * timestamp)
{
	struct qed_hwfn *hwfn = QED_LEADING_HWFN(cdev);

	return qed_ptp_read_rx_ts(hwfn, hwfn->p_ptp_ptt, timestamp);
}

static int qed_ptp_hw_read_tx_ts(struct qed_dev *cdev, u64 * timestamp)
{
	struct qed_hwfn *hwfn = QED_LEADING_HWFN(cdev);

	return qed_ptp_read_tx_ts(hwfn, hwfn->p_ptp_ptt, timestamp);
}

static int qed_ptp_hw_read_cc(struct qed_dev *cdev, u64 * cycles)
{
	struct qed_hwfn *hwfn = QED_LEADING_HWFN(cdev);

	return qed_ptp_read_cc(hwfn, hwfn->p_ptp_ptt, cycles);
}

static int qed_ptp_hw_disable(struct qed_dev *cdev)
{
	struct qed_hwfn *hwfn = QED_LEADING_HWFN(cdev);

	return qed_ptp_disable(hwfn, hwfn->p_ptp_ptt);
}

static int qed_ptp_hw_adjfreq(struct qed_dev *cdev, s32 ppb)
{
	struct qed_hwfn *hwfn = QED_LEADING_HWFN(cdev);

	return qed_ptp_adjfreq(hwfn, hwfn->p_ptp_ptt, ppb);
}

static int qed_ptp_hw_enable(struct qed_dev *cdev)
{
	struct qed_hwfn *hwfn = QED_LEADING_HWFN(cdev);

	return qed_ptp_enable(hwfn, hwfn->p_ptp_ptt);
}

const struct qed_eth_ptp_ops qed_ptp_ops_pass = {
	INIT_STRUCT_FIELD(hwtstamp_tx_on, qed_ptp_hw_hwtstamp_tx_on),
	INIT_STRUCT_FIELD(cfg_rx_filters, qed_ptp_hw_cfg_rx_filters),
	INIT_STRUCT_FIELD(read_rx_ts, qed_ptp_hw_read_rx_ts),
	INIT_STRUCT_FIELD(read_tx_ts, qed_ptp_hw_read_tx_ts),
	INIT_STRUCT_FIELD(read_cc, qed_ptp_hw_read_cc),
	INIT_STRUCT_FIELD(adjfreq, qed_ptp_hw_adjfreq),
	INIT_STRUCT_FIELD(disable, qed_ptp_hw_disable),
	INIT_STRUCT_FIELD(enable, qed_ptp_hw_enable),
};
